package sets;

// override all methods, but reuse head, tail fields
public class LockFreeSet<T> extends SequentialSet<T>
{
    public LockFreeSet() {
        tail = newNode(Integer.MAX_VALUE);
        head = newNode(Integer.MIN_VALUE, tail);
    }

    // does `node' have given `key' and is it valid?
    @Override
    protected boolean rawHas(Node<T> node, int key) {
        return super.rawHas(node, key) && node.valid();
    }

    // find without locking, skipping over invalid nodes, validate at the end
    @Override
    public boolean has(T item) {
        int key = item.hashCode();
        Node<T> pred = null, curr = null; // dummy initialization
        // first node from head whose key >= `key'
        Position<T> where = super.find(head, key);  // sequential find just goes through the list
        pred = where.pred;
        curr = where.curr;
        // pred.key < key <= curr.key
        return rawHas(curr, key);  // found if right content and not invalidated
    }

    // find while removing invalid nodes
    // if add fails because another thread interferes, repeat
    @Override
    public boolean add(T item) {
        do {
            // new node
            Node<T> pred = null, curr = null; // dummy initialization
            // first node from head whose key >= `node.key'
            Position<T> where = find(head, item.hashCode());
            pred = where.pred;
            curr = where.curr;
            // pred.key < node.key <= curr.key
            Node<T> node = newNode(item, curr);
            // if item already in set, return false
            if (super.rawHas(curr, node.key())) return false;
            // if pred still points to curr, make it point to node
            if (pred.setNextIfValid(curr, node)) return true;
        } while (true);  // repeat until stable outcome
    }

    // find while removing invalid nodes
    // if remove fails because another thread interferes, repeat
    @Override
    public boolean remove(T item) {
        int key = item.hashCode();
        Node<T> pred = null, curr = null; // dummy initialization
        do {
            // first node from head whose key >= `key'
            Position<T> where = find(head, key);
            pred = where.pred;
            curr = where.curr;
            // pred.key < key <= curr.key
            // if item not in set, return false
            if (!super.rawHas(curr, key)) return false;
            // if node is modified while trying to invalidate it: retry
            if (!curr.setInvalid()) continue;
            // try to remove curr by redirecting pred
            // if it fails, another thread will do it
            pred.setNextIfValid(curr, curr.next());
            return true;
        } while (true);
    }

    // find position while removing invalid nodes
    @Override
    protected Position<T> find(Node<T> start, int key) {
        Node<T> nextValid;
        boolean valid;
        // predecessor, current, and successor node in iteration
        Node<T> pred = null, curr = null, succ = null;
        retry: do {
            pred = start;        // from `start' node
            curr = pred.next();
            do {
                nextValid = curr.nextValid();
                succ = nextValid.next();
                valid = nextValid.valid();
                while (!valid) {  // curr is not valid: remove it
                    // set pred.next to succ
                    boolean success = pred.setNextIfValid(curr, succ);
                    // !success iff pred has been changed since the current thread read it
                    if (!success) continue retry;  // in this case, start from scratch
                    curr = succ;  // if removal of curr is successful, succ is the new curr
                    nextValid = curr.nextValid();
                    succ = nextValid.next();
                    valid = nextValid.valid();
                }
                // pred and curr are now valid
                // if pred and curr denote the correct position, return them
                if (curr.key() >= key) return new Position<T>(pred, curr);
                // otherwise, move to the next node
                pred = curr;
                curr = succ;
            } while (true);
        } while (true);
    }

    // node with mark and next atomically accessible
    @Override
    protected Node<T> newNode(T item) {
        return new LockFreeNode<>(item);
    }

    // node with mark and next atomically accessible
    @Override
    protected Node<T> newNode(int key) {
        return new LockFreeNode<>(key);
    }

    // new node pointing to succ
    protected Node<T> newNode(T item, Node<T> succ) {
        return new LockFreeNode<>(item, succ);
    }

    // new node pointing to succ
    protected Node<T> newNode(int key, Node<T> succ) {
        return new LockFreeNode<>(key, succ);
    }

}
